/*
   Projekt        : SAPHIR-Online-Monitoring 
   Name der Datei : $Id: dvmd_main.c,v 1.25 1997/01/29 17:08:03 hannappe Exp $
   Autor          : J. Hannappel
   Beschreibung   : Prozess zur Auslese der einfachen Digital-Voltmeter
            mit RS-232-Schnittstelle.
	    Das Programm startet zwei Prozesse, von denen der eine
	    andauenrd auf die angeschlossenen DVM's hoert und der
	    andere per RPC die Daten zur verfuegung stellt.
*/

#define FACILITY "OnLine: DVMd: "
     /*
                        Die Includes:
       ---------------------------------------------------------------------
     */
#include "dvm_access.h"  /* rpc-definitions */

#include <stdio.h>
#include <stdlib.h>
#include <rpc/pmap_clnt.h> /* for pmap_unset */
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <signal.h>
#include <syslog.h>
#include <errno.h>

     /*
                        Die Defines:
       ---------------------------------------------------------------------
     */
#define CHARS_PER_VALUE (13)
#define MAXDVMS (8)
#define MAXNAMELEN (128)
#define PIDFILE "/tmp/dvmd_reader_pid"
#define MAINPIDFILE "/tmp/dvmd_pid"

#define MAX(a,b) ((a)>(b)?(a):(b))
/*
   Deklaration von Datentypen:
   ---------------------------------------------------------------------
   */

struct dvm_type 
{
  int fd;		/* file descriptor of device file, -1 means unused */
  int config_flag;      /* used by config function ONLY */
  int tm_value_so_skip_return; /* flag used to overcome DVM bug */
  char device[MAXNAMELEN];	/* name of device file */
  char name[MAXNAMELEN];	/* Arbitrary name of connected device */
  char *wp;             /* write pointer to value buffer */
  struct timeval first_time; /* time of receipt of first character
				of the currently building value */
  struct timeval last_time; /* time of start of receipt of last complete
			       value */
  struct timeval last_char; /* time of  receipt of last character */
  char value[16];	/* last complete value */
  char value_buffer[16];/* buffer to store incoming characters */
};

     /*
                        statische Variablen:
	       (durch grossen Anfangsbuchstaben gekennzeichnet!)
       ---------------------------------------------------------------------
     */

static struct dvm_type *Dvm;
static char *Config_file;

static int Shm_id = -1; /* Id of shared memory segment */
static pid_t Serial_process_pid = 0; /* PID of Serial read subprocess */ 
     /*
              Protoypen der 'static' deklarierten Funktionen:
       ---------------------------------------------------------------------
     */
static int init_dvm_lines(const char *config_file);
static int open_dvm_line(const char *device);
static int read_character(struct dvm_type *dvm);
static void do_read_loop(void); 
static int handle_rpc_calls(void);
static void init_daemon(int *argc, char *argv[]);

static void reconfigure(int dummy);
static void remove_pid_file(void);
static void remove_mainpid_file(void);
static void exit_handler(void);

static void termination_handler(int signal);
static void termination_handler2(int signal);

/* from dvm_acces_svc.c */
void dvm_read_prog_1(struct svc_req *rqstp, register SVCXPRT *transp);

    /*
                         Definitionsteil:
       ---------------------------------------------------------------------
     */
int 
main(int argc, char *argv[])
{
  init_daemon(&argc,argv);
  openlog("dvmd",LOG_CONS | LOG_NDELAY | LOG_PID, LOG_LOCAL0);
  syslog(LOG_NOTICE,"starting $Id: dvmd_main.c,v 1.25 1997/01/29 17:08:03 hannappe Exp $");
  /* create shared memory segment */
  Shm_id = shmget(IPC_PRIVATE,MAXDVMS*sizeof(struct dvm_type),
		  IPC_CREAT | IPC_EXCL | 0644);
  if (Shm_id == -1)
    {
      syslog(LOG_ERR,"shmget failed %m");
      exit(1);
    }
  /* and attach to it (map into user space) */
  Dvm = (struct dvm_type *)shmat(Shm_id,NULL,0);
  if (Dvm == (void *)-1)
    {
      syslog(LOG_ERR,"shmat failed %m");
      exit(1);
    }
  shmctl(Shm_id,IPC_RMID,NULL); /* destroy after last detach */
  atexit(exit_handler); /* install exit handler to kill subprocesses */
  if (signal(SIGTERM,termination_handler) == SIG_ERR)
    syslog(LOG_ERR,"Failed to install SIGTERM handler: %m");
  if (signal(SIGCHLD,termination_handler) == SIG_ERR)
    syslog(LOG_ERR,"Failed to install SIGCHLD handler: %m");

  { /* now mark all dvm's as unused by setting the FD to -1 */
    int i;
    for (i=0; i< MAXDVMS; i++)
      Dvm[i].fd = -1;
  }

  if (argc == 2)
    {
      Config_file = malloc(strlen(argv[1])+1);
      if (!Config_file)
	{
	  syslog(LOG_ERR,"failed to allocate memory: %m");
	  exit(1);
	}
      strcpy(Config_file,argv[1]);
    }
  else
    Config_file = "/etc/dvmd.config";
  if (!init_dvm_lines(Config_file))
    exit(1);
  
  { /* spawn subprocess for reading dmm values */
    Serial_process_pid = fork();
    if (Serial_process_pid == 0)
      {	/* now we are the forked process */
	FILE *pid_file;
	syslog(LOG_NOTICE,"starting read subprocess");
	strcpy(argv[0],"dvmd reader"); /* set process name */
	pid_file = fopen(PIDFILE,"w");
	if (!pid_file)
	  syslog(LOG_ERR,"can't open pid file " PIDFILE " %m");
	fprintf(pid_file,"%d\n",getpid());
	fclose(pid_file);
	atexit(remove_pid_file);
	if (signal(SIGTERM,termination_handler2) == SIG_ERR)
	  syslog(LOG_ERR,"Failed to install SIGTERM handler: %m");
	if (signal(SIGHUP,reconfigure) == SIG_ERR)
	  syslog(LOG_ERR,"Failed to install SIGHUP handler: %m");
	do_read_loop();
      }
    else if (Serial_process_pid != -1)
      {	/* now we are still the old process */
	handle_rpc_calls();
      }
    else
      {	/* dirt! the fork failed! */
	syslog(LOG_ERR,"Fork failed!");
	exit(1);
      }
  }
  return(0);
}

     /*
       ---------------------------------------------------------------------
       */
static void 
init_daemon(int *argc, char *argv[])
{
  {
    FILE *pid_file;
    pid_t pid;
    pid_file = fopen(MAINPIDFILE,"r");
    if (pid_file)
      {
	fscanf(pid_file,"%d",&pid);
	fprintf(stderr,"Killing old daemon(process %d)...",pid);
	fclose(pid_file);
	if (kill(pid,SIGTERM) != 0)
	  {
	    if (errno != ESRCH)
	      {
		perror("could not kill old daemon");
		exit(1);
	      }
	    else
	      {
		fprintf(stderr," seems to be dead already,\n"
			"pidfile \"%s\" probably  left over from crash,"
			" deleting it.\n",
			MAINPIDFILE);
		if (unlink(MAINPIDFILE) != 0)
		  {
		    perror("remove of pid file failed");
		    fprintf(stderr,
			    "clean up please, and remove \"%s\" also!\n",
			    PIDFILE);
		    exit(1);
		  }
		if (unlink(PIDFILE) !=0)
		  if (errno != ENOENT)
		    {
		      perror("remove of pid file failed");
		      fprintf(stderr,"please remove \"%s\"!\n",PIDFILE);
		      exit(1);
		    }
		  else
		    fprintf(stderr,
			    "strange, the subprocess pidfile didn't exist.\n");
	      }
	  }
	{
	  int retries=20;
	  int status;
	  for (;;)
	    {
	      status=kill(pid,SIGTERM);
	      if (status == -1 && errno == ESRCH) /* process is dead, go on */
		break;
	      sleep(1);
	      fprintf(stderr,".");
	      retries--;
	      if (!retries)
		{
		  perror("\nsubprocess didn't die");
		  exit(1);
		}
	    }
	  fprintf(stderr,"done.\n");
	}
      }
  }

  {
    int daemon_pid;
    daemon_pid = fork();
    if (daemon_pid == 0)
      {	/* now we are the forked process */
	FILE *pid_file;
	pid_file = fopen(MAINPIDFILE,"w");
	if (!pid_file)
	  syslog(LOG_ERR,"can't open pid file " MAINPIDFILE " %m");
	fprintf(pid_file,"%d\n",getpid());
	fclose(pid_file);
	atexit(remove_mainpid_file);
	return;
      }
    else if (daemon_pid != -1)
      {	/* now we are still the old process */
	exit(0);
      }
    else
      {	/* dirt! the fork failed! */
	perror("dvmd: could not fork daemon");
	exit(1);
      }
  }
}
     /*
       ---------------------------------------------------------------------
       This function reads the config file and opens the DVM lines
       */
static int
init_dvm_lines(const char *config_file)
{
  FILE *file;
  
  file = fopen(config_file,"r");
  if (!file)
    {
      syslog(LOG_ERR,"failed to open config file %s: %m",config_file);
      return(FALSE);
    }
  { /* mark all DVMs as not configured  */
    int i;
    for (i=0; i < MAXDVMS; i++)
      {
	Dvm[i].config_flag = FALSE;
	Dvm[i].tm_value_so_skip_return = FALSE;
      }
  }

  syslog(LOG_INFO,"reading config file %s",config_file);
  while (!feof(file))
    {
      char line[1024];
      char *name;
      char *device;
      if (fgets(line,sizeof(line),file)==NULL)
	{
	  if (feof(file)) break;
	  syslog(LOG_ERR,"read error on config file %s: %m",config_file);
	  return(FALSE);
	}      
      if (line[0] == '#') continue; /* skip comment lines */
      { /* extract device and name from line */
	char *cp;
	size_t length;
	cp = line + strspn(line," \t");
	length = strcspn(cp," \t");
	device = cp;
	cp[length]='\0';
	if (length+1 > MAXNAMELEN)
	  {
	    syslog(LOG_ERR,
		   "overlong device name %s, must not exceed %d characters",
		   device,
		   MAXNAMELEN);
	    continue;
	  }
	cp+= length+1;
	cp += strspn(cp," \t");
	length = strcspn(cp,"\n");
	if (length+1 > MAXNAMELEN)
	  {
	    syslog(LOG_ERR,
		   "overlong symbolic name %s, must not exceed %d characters",
		   device,
		   MAXNAMELEN);
	    continue;
	  }
	name = cp;
	name[length]='\0';
	if (strstr(name,"unused") == name)
	  {
	    syslog(LOG_INFO,"the device %s is unused",device);
	    continue;
	  }
	else if (strstr(name,"used for") == name)
	  {
	    syslog(LOG_INFO,"the device %s is %s",device,name);
	    continue;
	  }
	else if (strcmp(name,"broken") == 0)
	  {
	    syslog(LOG_INFO,"the device %s is broken",device);
	    continue;
	  }
      }
      
      {
	int fd;
	int i;
	for (i=0; i < MAXDVMS; i++)
	  if (Dvm[i].fd != -1 &&
	      strcmp(name,Dvm[i].name) == 0 &&
	      strcmp(device,Dvm[i].device) == 0)
	    break;
	if (i < MAXDVMS) /* this configuration is unchanged */ 
	  {
	    syslog(LOG_INFO,
		   "the device %s remains \"%s\"",
		   device,name);
	    Dvm[i].config_flag = TRUE;
	    continue; 
	  }
	for (i=0; i < MAXDVMS; i++)
	  if (Dvm[i].fd != -1 &&
	      strcmp(device,Dvm[i].device) == 0)
	    break;
	if (i < MAXDVMS) /* only the name has changed */
	  {
	    char old_name[MAXNAMELEN];
	    strcpy(old_name,Dvm[i].name);
	    strcpy(Dvm[i].name,name);
	    syslog(LOG_INFO,
		   "the name attached to %s changed from \"%s\" to \"%s\"",
		   device,old_name,name);
	    Dvm[i].config_flag = TRUE;
	    continue;
	  }
	for (i=0; i < MAXDVMS; i++)
	  if (Dvm[i].fd == -1) break; /* -1 denotes an unused entry... */
	if (i >= MAXDVMS)
	  {
	    syslog(LOG_ERR,
		   "Can't handle more than %d DVMs at once, please enlarge me!\n",
		   MAXDVMS);
	    exit(1);
	  }
	fd = open_dvm_line(device);
	if (fd == -1) return(1);
	Dvm[i].fd = fd;
	flock(Dvm[i].fd,LOCK_EX);
	Dvm[i].config_flag = TRUE;
	strcpy(Dvm[i].device,device);
	strcpy(Dvm[i].name,name);
	Dvm[i].wp = &Dvm[i].value_buffer[0];
	timerclear(&(Dvm[i].last_time));
	timerclear(&(Dvm[i].last_char));
	syslog(LOG_INFO,"opened \"%s\" on device %s",
	       Dvm[i].name,
	       Dvm[i].device);
      }
    }
  fclose(file);
  { /* now close all unconfigured DVMs  */
    int i;
    for (i=0; i < MAXDVMS; i++)
      if (!Dvm[i].config_flag && Dvm[i].fd != -1)
	{
	  flock(Dvm[i].fd,LOCK_UN);
	  close(Dvm[i].fd);
	  Dvm[i].fd = -1;
	  syslog(LOG_INFO,"closed \"%s\" on device %s",
		 Dvm[i].name,
		 Dvm[i].device);
	}
  }
  return(TRUE);
}

     /*
       ---------------------------------------------------------------------
This function opens the device file coresponding to a DVM and
sets all line parameters to enable the DVM to talk over the
line. It returns the fd 
     */
static int
open_dvm_line(const char *device)
{
  int fd;
  struct termios fd_attr;

  fd = open(device,O_RDWR | O_NOCTTY);
  if (fd == -1)
    {
      syslog(LOG_ERR,"failed to open %s: %m",device);
      return(fd);
    }
  
  fd_attr.c_iflag = IGNBRK | IGNPAR;
  fd_attr.c_oflag = 0;
  fd_attr.c_cflag = CS7	/* seven bits per character */
    | CSTOPB		/* 2 stop bits */
      | CREAD		/* enable receiver */
	| CLOCAL;	/* ignore modem control lines */
  fd_attr.c_lflag = 0;
  cfsetospeed(&fd_attr,B1200); /* baud rate is 1200 Baud */
  cfsetispeed(&fd_attr,B1200);
  if (tcsetattr(fd,TCSANOW,&fd_attr) == -1) /* commit changes NOW */
    {
      syslog(LOG_ERR,"failed to tcsetattr on %s: %m",device);
      close(fd);
      fd=-1;
      return(fd);
    }
  {	/* lower RTS to provide power to the opto-coupler of the DMM */
    int modelines = 0;
    if (ioctl(fd,TIOCMGET,&modelines) == -1)
      {
	syslog(LOG_ERR,"failed to ioctl TIOCMGET on %s: %m",device);
	close(fd);
	fd=-1;
	return(fd);
      }
    modelines &= ~TIOCM_RTS;
    if (ioctl(fd,TIOCMSET,&modelines) == -1)
      {
	syslog(LOG_ERR,"failed to ioctl TIOCMSET on %s: %m",device);
	close(fd);
	fd=-1;
	return(fd);
      }
  }
  return(fd);
}
     /*
       ---------------------------------------------------------------------
       This function reads one caharcter on the dvm's file and
       assembles the value.
       If the value is read completely it is tranferred into the 
       Buffer for complete values and the buffer is cleared. 
       Beware! The DVM's show the end of a value by sending '\r',
       but due to a bug the '\r' is not sent if measuring 
       room temperature!
     */
static int 
read_character(struct dvm_type *dvm)
{
  gettimeofday(&(dvm->last_char),NULL);
  {
    int bytes;
    bytes = read(dvm->fd,dvm->wp,1);
    if (bytes != 1)
      {
	syslog(LOG_DEBUG,"read error on %s: %m",dvm->device);
	return(FALSE);
      }
  }
  if (*dvm->wp == '\r') /* the end of a complete value */
    {
      if (dvm->tm_value_so_skip_return)
	dvm->tm_value_so_skip_return = FALSE; /* reset to default value */
      else if (dvm->wp - &dvm->value_buffer[0] == CHARS_PER_VALUE)
	{ /* a complete value has been received! */
	  memcpy(dvm->value,dvm->value_buffer,CHARS_PER_VALUE);
	  dvm->value[CHARS_PER_VALUE] = '\0';
	  dvm->last_time = dvm->first_time;
	}
      else
	{
	  *dvm->wp = '\0';
	  syslog(LOG_DEBUG,"incomplete value on %s: %s",
		 dvm->device,dvm->value_buffer);
	}
      dvm->wp = &dvm->value_buffer[0];
      return(TRUE);
    }
  else if (*dvm->wp == 'C') /* possibly the end of a room temperature */
    {
      if (dvm->wp - &dvm->value_buffer[0] == CHARS_PER_VALUE  &&
	  dvm->value_buffer[0] == ' ' && 
	  dvm->value_buffer[1] == 'T' && dvm->value_buffer[2] == 'M')
	{ /* a complete Temerature value has been received! */
	  memcpy(dvm->value,&dvm->value_buffer[1],CHARS_PER_VALUE);
	  dvm->value[CHARS_PER_VALUE] = '\0';
	  dvm->last_time = dvm->first_time;
	  dvm->tm_value_so_skip_return = TRUE; /* if NOT a buggy DVM or
						not room temp, just
						ignore the following '\r' */
	  dvm->wp = &dvm->value_buffer[0];
	  return(TRUE);
	}
      else if (dvm->wp > &dvm->value_buffer[0] + CHARS_PER_VALUE)
	{
	  *dvm->wp = '\0';
	  syslog(LOG_DEBUG,"incomplete value on %s: \"%s\" %d",
		 dvm->device,dvm->value_buffer,dvm->wp - &dvm->value_buffer[0]);
	  dvm->wp = &dvm->value_buffer[0];
	  return(TRUE);
	}
    }
  if (dvm->wp == &dvm->value_buffer[0])
    gettimeofday(&(dvm->first_time),NULL);
  dvm->wp++;
  if (dvm->wp >= &dvm->value_buffer[sizeof(dvm->value_buffer)])
    { /* somehow we must have missed a '\r', so clear buffer and try anew */
      dvm->value_buffer[sizeof(dvm->value_buffer)-1] = '\0';
      syslog(LOG_DEBUG,"overlong value on %s: %s",
	     dvm->device,dvm->value_buffer);
      dvm->wp = &dvm->value_buffer[0]; 
    }
  return(TRUE);
}

     /*
       ---------------------------------------------------------------------
       this function performs the read loop on all attached DVM's 
       and calls the value assemblage functions 
       */
static void
do_read_loop(void)
{ 
  for (;;)
    { /* wait on all open DMMs for a character and assemble the
	 characters into the value buffers */
      struct timeval timeout;
      struct timeval now;
      int nof_ready;
      fd_set read_fds;
      int max_fd=0;
      
      FD_ZERO(&read_fds);
      
      {
	int i;
	for (i=0; i<MAXDVMS; i++)
	  {
	    if (Dvm[i].fd == -1) continue; /* skip unused entrys */
	    max_fd = MAX(Dvm[i].fd,max_fd);
	    FD_SET(Dvm[i].fd,&read_fds);
	  }
      }
      timeout.tv_sec = 0;
      timeout.tv_usec = 900000;
      nof_ready = select(max_fd+1,&read_fds,NULL,NULL,&timeout);
      gettimeofday(&now,NULL);
      if (nof_ready < 0) /* error occured */
	{
	  if (errno == EINTR) /* interrupted system call,
				 probably a SIGHUP */
	    {
	      syslog(LOG_DEBUG,"select interrupted");
	      continue;
	    }
	  syslog(LOG_ERR,"select failed: %m");
	  exit(1);
	}
      {
	int i;
	for (i=0; i<MAXDVMS; i++)
	  {
	    double sincelast;
	    if (Dvm[i].fd == -1) continue; /* skip unused entrys */
	    sincelast = now.tv_sec - Dvm[i].last_char.tv_sec;
	    sincelast += (now.tv_usec - Dvm[i].last_char.tv_usec) * 1e-6;

	    if (FD_ISSET(Dvm[i].fd,&read_fds))
	      { /* read a single character */
		read_character(&(Dvm[i]));
	      }
	    else if (sincelast > 0.9)
	      { /* no character arrived since at least a second,
		   so probably the Dvm isn't in com-mode.
		   Send it a "D" to make it reply with the value in the display
		   set also the last_char time to keep from sending "D"'s
		   to fast!
		   We dont have to worry about blocking,
		   the characters are sent out even if no listener exists */
		write(Dvm[i].fd,"D",1);
		memcpy(&Dvm[i].last_char,&now,sizeof(now));
	      }
	  }
      }
    }
}

     /*
       ---------------------------------------------------------------------
       */
static int 
handle_rpc_calls(void)
{
  SVCXPRT *transp;

  (void) pmap_unset(DVM_READ_PROG, DVM_READ_PROG);

  transp = svcudp_create(RPC_ANYSOCK);
  if (transp == NULL)
    {
      syslog(LOG_ERR,"cannot create udp service: %m");
      exit(1);
    }
  if (!svc_register(transp, DVM_READ_PROG, DVM_READ_PROG, 
		    dvm_read_prog_1, IPPROTO_UDP)) 
    {
      syslog(LOG_ERR,"unable to register (DVM_READ_PROG, DVM_READ_PROG, udp). %m");
      exit(1);
    }

  transp = svctcp_create(RPC_ANYSOCK, 0, 0);
  if (transp == NULL)
    {
      syslog(LOG_ERR,"cannot create tcp service: %m");
      exit(1);
    }
  if (!svc_register(transp, DVM_READ_PROG, DVM_READ_PROG,
		    dvm_read_prog_1, IPPROTO_TCP)) {
      syslog(LOG_ERR,"unable to register (DVM_READ_PROG, DVM_READ_PROG, tcp). %m");
    exit(1);
  }

  svc_run();
  syslog(LOG_ERR,"svc_run returned! errno is: %m");
  exit(1);
  /* NOTREACHED */
}

     /*
       ---------------------------------------------------------------------
       */
dvm_value_type * 
dvm_read_1_svc(char *name,  struct svc_req *rqstp)
{
  
  static dvm_value_type  result;
  int i;
  result.value=0.0;
  strcpy(result.unit,"fail");
  result.tv_sec=result.tv_usec=0;
  for (i=0; i< MAXDVMS; i++)
    {
      if (Dvm[i].fd == -1) continue; /* skip unused entrys */
      if (strcmp(Dvm[i].name,name)==0)
	{
	  char value[16];
	  result.tv_sec = Dvm[i].last_time.tv_sec;
	  result.tv_usec = Dvm[i].last_time.tv_usec;
	  memcpy(value,Dvm[i].value,16);
	  sscanf(value,"%*2c%f %s",
		 &result.value,
		 result.unit);
	  break;
	}
    }
  return(&result);
}

     /*
       ---------------------------------------------------------------------
       The exit handler removes the shared memory and 
       also kills the suprocesses
       */
static void 
exit_handler(void)
{
  if (Serial_process_pid != 0)
    {
      int status;
      if (kill(Serial_process_pid,SIGTERM) == -1)
	{
	  if (errno == ESRCH)
	    syslog(LOG_NOTICE,"suprocesses already gone");
	  else
	    syslog(LOG_NOTICE,"could not kill subprocess: %m");
	}
      else
	{
	  if (waitpid(Serial_process_pid,&status,0) != Serial_process_pid)
	    syslog(LOG_ERR,"subprocess didn't die! %m");
	  else
	    syslog(LOG_NOTICE,"subprocess exited");
	}
    }
}
     /*
       ---------------------------------------------------------------------
       */
static void 
remove_pid_file(void)
{
  syslog(LOG_NOTICE,"exiting");
  if (unlink(PIDFILE) != 0)
    syslog(LOG_ERR,"can't remove " PIDFILE ":%m");
}
     /*
       ---------------------------------------------------------------------
       */
static void 
remove_mainpid_file(void)
{
  if (unlink(MAINPIDFILE) != 0)
    syslog(LOG_ERR,"can't remove " MAINPIDFILE ":%m");
}
     /*
       ---------------------------------------------------------------------
       This function re-inits the dvm-lines from the config file.
       It is to be invoked by a SIGHUP on the read process!
       */
static void 
reconfigure(int dummy)
{
  syslog(LOG_INFO,"SIGHUP caught, re-configuring");
  (void)init_dvm_lines(Config_file)  ;
  if (signal(SIGHUP,reconfigure) == SIG_ERR)
    syslog(LOG_ERR,"Failed to re-install SIGHUP handler: %m");
}
     /*
       ---------------------------------------------------------------------
       */
static void
termination_handler(int signal)
{
  switch (signal) {
  case SIGTERM: 
    syslog(LOG_INFO,"SIGTERM caught, finishing"); 
    exit_handler();
  case SIGCHLD:
    syslog(LOG_INFO,"SIGCHLD caught (subprocess died), finishing");
    break;
  default:
    syslog(LOG_INFO,"unconfigured signal %d caught, finishing",signal);
  }
  remove_mainpid_file();
  _exit(0);
}

     /*
       ---------------------------------------------------------------------
       */
static void
termination_handler2(int signal)
{
  syslog(LOG_INFO,"SIGTERM caught, finishing");
  remove_pid_file();
  _exit(0);
}




