/*
   Projekt        : Online-Ueberwachung
   Name der Datei : $Id: mux_reader_main.c,v 1.5 1997/04/02 20:22:55 hannappe Exp $
   Autor          : J.Hannappel
   Beschreibung   : Auslesen gemultiplexter DVMs
 */

#define FACILITY "Mux-Reader: "
     /*
        Die Includes:
        ---------------------------------------------------------------------
      */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <float.h>

#include "dvm_access.h"

     /*
        Die Defines:
        ---------------------------------------------------------------------
      */

     /*
        Deklaration von Datentypen:
        ---------------------------------------------------------------------
      */
struct channel_data_type
  {
    double last_readout_time;
    float value;
    unsigned char channel_no; /* must be unsigned char to work
				 with the mux-driver */
  };
     /*
        globale Variablen:
        (durch 3-Grossbuchstaben-Code gekennzeichnet!)
        ---------------------------------------------------------------------
      */

static struct channel_data_type *Channel = NULL;
static int Nof_channels = 0;
static int Nof_cycles = 0;
static int T_switch_to_read = 3;
static int T_read_to_switch = 1;
static int T_inter_cycle = 10;
static double Min_t_switch_to_read = 1.2;
static FILE *Output = NULL;
static const char *Mux_device = "/dev/muxbits";
static int Verbose = 0;
     /*
        statische Variablen:
        (durch grossen Anfangsbuchstaben gekennzeichnet!)
        ---------------------------------------------------------------------
      */


     /*
        Protoypen der 'static' deklarierten Funktionen:
        ---------------------------------------------------------------------
      */
static int do_read_loop(CLIENT *clnt, char *name);
static void dvm_read_prog_1 (char *host, char *name);
     /*
        Definitionsteil:
        ---------------------------------------------------------------------
      */

     /*
        ---------------------------------------------------------------------
      */

     /*
        ---------------------------------------------------------------------
      */

static int
do_read_loop(CLIENT *clnt, char *name)
{
  int cycle;
  int mux_fd;
  mux_fd = open(Mux_device, O_WRONLY | O_SYNC);
  if (mux_fd == -1)
    {
      perror("Can't open Mux device"); 
      return(0);
    }
  for (cycle=1; Nof_cycles==0 || cycle < Nof_cycles;cycle++)
    {
      int i;
      double t_min, t_max;
      t_min = DBL_MAX;
      t_max = 0;
      for (i=0; i<Nof_channels; i++)
	{
	  struct timeval now;
	  double switch_time;
	  double read_time;
	  dvm_value_type *result_1;
	  
	  write(mux_fd,&Channel[i].channel_no,sizeof(Channel[i].channel_no));
	  gettimeofday(&now,NULL);
	    switch_time = now.tv_sec + 1.0E-6*now.tv_usec;
	    sleep(T_switch_to_read);
	    result_1 = dvm_read_1 (name, clnt);
	    if (result_1 == NULL)
	      {
		clnt_perror (clnt, "call failed:");
		continue;
	      }
	    read_time = result_1->tv_sec + 1.0E-6*result_1->tv_usec;
	    if (read_time < t_min) t_min = read_time;
	    if (read_time > t_max) t_max = read_time;
	    if (read_time - switch_time < Min_t_switch_to_read)
	      {
		fprintf(stderr,"timing error, t_switch=%13.3f t_read=%13.3f\n",
			switch_time, read_time);
		sleep(T_read_to_switch);
		continue;
	      }
	    Channel[i].last_readout_time = read_time;
	    Channel[i].value = result_1->value;
	    sleep(T_read_to_switch);
	}
      fprintf(Output,"%d\t%d", (int)t_min, (int)t_max);
      for (i=0; i<Nof_channels; i++)
	fprintf(Output,"\t%g",Channel[i].value);
      fprintf(Output,"\n");
      fflush(Output);
      sleep(T_inter_cycle);
    }
  return(-1);
}

void
dvm_read_prog_1 (char *host, char *name)
{
  CLIENT *clnt;
  clnt = clnt_create (host, DVM_READ_PROG, DVM_READ_PROG, "udp");
  if (clnt == NULL)
    {
      clnt_pcreateerror (host);
      exit (1);
    }
  do_read_loop(clnt, name);
  clnt_destroy (clnt);
}

int
main (int argc, char *argv[])
{
  char *host = "localhost";
  char *name = "muxdvm";
  char *output_file = NULL;
  for (;;)
    {
      int c;
      c = getopt(argc, argv, "vo:d:n:h:r:s:c:k:");
      switch(c)
	{
	case -1: break;
	case 'v': Verbose = -1; break;
	case 'o': output_file = optarg; break;
	case 'd': Mux_device = optarg; break;
	case 'n': name = optarg; break;
	case 'h': host = optarg; break;
	case 'r': T_switch_to_read = atoi(optarg); break;
	case 's': T_read_to_switch = atoi(optarg); break;
	case 'c': T_inter_cycle = atoi(optarg); break;
	case 'k': Nof_cycles = atoi(optarg); break;
	case ':':
	case '?':
	default :
	  fprintf(stderr,"Usage: %s <options> channels...\n"
		  "\t-v\tverbose\n"
		  "\t-o <file>\toutput file name [%s]\n"
		  "\t-d <file>\tmux device file name [%s]\n"
		  "\t-h <host>\tdvmd host name [%s]\n"
		  "\t-n <name>\tdvm name for dvmd [%s]\n"
		  "\t-r <int>\ttime from switch to read, seconds [%d]\n"
		  "\t-s <int>\ttime from read to switch, seconds [%d]\n"
		  "\t-c <int>\ttime between cycles, seconds [%d]\n"
		  "\t-k <int>\tnumber of cycles, 0=inf [%d]\n",
		  argv[0],
		  output_file, Mux_device, host, name,
		  T_switch_to_read, T_read_to_switch, T_inter_cycle,
		  Nof_cycles);
	  exit(1);
	}
      if (c == -1) break;
    }
  if (output_file)
    Output = fopen(output_file,"w");
  else
    Output = stdout;

  {
    int i;
    for (i=optind; i < argc; i++)
      Nof_channels++;
    Channel = malloc(sizeof(*Channel) * Nof_channels);
    for (i=optind; i < argc; i++)
      {
	int c;
	memset(&Channel[i-optind],0,sizeof(*Channel));
	c = atoi(argv[i])-1;
	if (c < 0)
	  {
	    fprintf(stderr,"channel numbers must be positive\n");
	    exit;
	  }
	Channel[i-optind].channel_no = c; 
	if (Verbose)
	  fprintf(stderr,"the %d-th channel is number %d\n",
		  i-optind+1, Channel[i-optind].channel_no+1);
      }
  }

  dvm_read_prog_1 (host, name);
  return (0);
}
